"""The redis storage engine."""
from .._daemon import Daemon


class Redis(Daemon):
    """A redis client and server."""

    _command = 'redis-server'
    _config_base = 'redis.conf'
    _config = dict(
        ipaddr='127.0.0.1',
        port='8776',
    )

    async def _prepare(self, config):
        """Get the configurate file ready."""
        self._command.append(str(config))



"""Unit tests for spools."""
import trio

from reel import Spool


async def test_spool_send_chunk():
    """A spool can send its output one chunk at a time."""
    now = Spool("date '+%Y-%m-%d'")
    async with now:
        async with trio.open_nursery() as nursery:
            now.start(nursery)
            assert now.pid



"""Test for the Track class."""
import logging

import trio

from reel import (
    Spool, Streamer, Track, Transport
)

LOG = logging.getLogger(__name__)


async def test_init_echo():
    """Create a Track object that echoes the input to output."""
    echo = Track(lambda _: _)
    assert isinstance(echo, Track)


async def test_start():
    """Start a Track with start() from the Streamer interface."""
    echo = Track(lambda _: _)
    assert isinstance(echo, Streamer)
    async with echo:
        async with trio.open_nursery() as nursery:
            echo.start(nursery)
            nursery.start_soon(echo.send_all, b'asdf')
            assert await echo.receive_some() == b'asdf'
            assert await echo.receive_some() is None


async def test_in_transport():
    """Run a Track in a Transport."""
    async with Track(lambda _: _) as echo:
        assert isinstance(echo, Transport)
        assert await echo.read('asdf') == 'asdf'


async def test_stream_through_track():
    """Stream some data through a Track."""
    async with Spool('cat') | Track(lambda _: _.upper()) as echo_up:
        assert await echo_up.read('this') == 'THIS'



"""Transport class."""
import logging

import trio

LOG = logging.getLogger(__name__)


class Transport(trio.abc.AsyncResource):
    """A device for running spools."""

    def __init__(self, *args):
        """Create a transport chain from a list of spools."""
        self._cancel_scope = None
        self._is_done = trio.Event()
        self._nursery = None
        self._output = None
        if len(args) == 1 and isinstance(args[0], list):
            self._chain = []
            for spool in args[0]:
                self._chain.append(spool)
        else:
            self._chain = list(args)
        LOG.debug(self.__repr__())

    def __gt__(self, nursery):
        """Set the nursery to use in the context manager."""
        self._nursery = nursery
        return self

    def __or__(self, next_one):
        """Store all the chained spools and reels in this transport."""
        self._chain.append(next_one)
        return self

    def __rshift__(self, next_one):
        """Store all the chained spools and reels in this transport."""
        self._chain.append(next_one)
        return self

    async def aclose(self):
        """Clean up resources."""
        for streamer in self._chain:
            await streamer.aclose()

    def __str__(self):
        """Print prettily."""
        return self.__repr__()

    def __repr__(self):
        """Represent prettily."""
        result = 'Transport([ '
        for thing in self._chain:
            result += thing.__repr__()
            result += ','
        result += ' ])'
        return result

    @property
    def is_done(self):
        """Has this thing finished playing."""
        return self._is_done.is_set()

    async def wait(self):
        """Wait until this is done."""
        await self._is_done.wait()

    async def _run(self, nursery, message=None):
        """Connect the spools with pipes and let the bytes flow."""
        for idx, spool in enumerate(self._chain):
            if idx == 0:  # Gets stdin
                spool.start(nursery, message)
            else:
                spool.start(nursery)

                # Create a pipe
                _src = self._chain[idx - 1]
                _dst = spool
                send_ch, receive_ch = trio.open_memory_channel(0)
                async with send_ch, receive_ch:
                    nursery.start_soon(
                        _src.send_to_channel, send_ch.clone()
                    )
                    nursery.start_soon(
                        _dst.receive_from_channel, receive_ch.clone()
                    )

        # Read stdout from the last spool in the list
        LOG.debug('about to read stdout of chain')
        ch_send, ch_receive = trio.open_memory_channel(0)
        nursery.start_soon(self._chain[-1].send_to_channel, ch_send)
        async for chunk in ch_receive:
            if not self._output:
                self._output = b''
            self._output += chunk

        self._is_done.set()
        if self._cancel_scope:
            self._cancel_scope.cancel()

    def spawn_in(self, nursery):
        """Start this transport and return a cancel_scope."""
        nursery.start_soon(self._run, nursery)
        self._cancel_scope = trio.CancelScope()
        return self._cancel_scope

    def start_daemon(self, nursery):
        """Run this transport in the background and return."""
        nursery.start_soon(self._run, nursery)

#    async def stop(self):
#        """Stop this transport."""
#        LOG.debug('{ stop %s }', self.__repr__())
#        for streamer in self._chain:
#            await streamer.stop()

    async def play(self) -> None:
        """Run this transport and ignore stdout."""
        async with trio.open_nursery() as nursery:
            await self._run(nursery)

    async def read(self, message=None, text=True):
        """Run transport and return stdout as str."""
        if message and text:
            message = message.encode('utf-8')

        async with trio.open_nursery() as nursery:
            await self._run(nursery, message=message)

        if text and self._output:
            decoded = self._output.decode('utf-8')

            # Remove trailing \n
            if decoded[-1:] == '\n':
                decoded = decoded[:-1]
            return decoded
        return self._output

    async def readlines(self):
        """Run this transport and return stdout as a `list`."""
        async with trio.open_nursery() as nursery:
            await self._run(nursery)
        if self._output:
            result = self._output
        else:
            result = b''
        output = result.decode('utf-8')
        return output.split('\n')[:-1]



"""Tests for the reel.Reel class."""
import logging

import pytest
import trio

from reel import Reel, Spool
from reel.cmd import ffmpeg

LOG = logging.getLogger(__name__)

# pylint: disable=not-async-context-manager


async def test_kill_old_processes(audio_dest):
    """Terminate each process when it finishes streaming."""

    async def checkpoint(spool):
        """Make sure we don't accumulate ffmpeg processes."""
        assert spool
        async with Spool('ps ax') | Spool('grep ffmpeg') as proclist:
            proc_count = len(await proclist.readlines())
            if checkpoint.count == 0:
                checkpoint.count = proc_count
            assert proc_count <= checkpoint.count + 1
    checkpoint.count = 0

    playlist = Reel([ffmpeg.read(track) for track in [
        '/Users/zach/out000.wav', '/Users/zach/out001.wav',
        '/Users/zach/out002.wav', '/Users/zach/out003.wav',
        '/Users/zach/out004.wav', '/Users/zach/out005.wav',
    ]], a_announce_to=checkpoint)
    async with playlist | audio_dest() as player:
        await player.play()
        assert player


async def test_reel_has_next_track(neil_reel):
    """A reel has a reference to the next track in the reel."""

    # A reel does not have track info until started.
    assert not neil_reel.current_track
    assert not neil_reel.next_track

    # Calculate the expected sequential playlist.
    expected = {}
    for idx, track in enumerate(neil_reel.tracks):
        if idx + 1 < len(neil_reel.tracks):
            expected[track] = neil_reel.tracks[idx + 1]
        elif idx + 1 == len(neil_reel.tracks):
            expected[track] = None

    # Confirm current/next tracks at each announcement.
    def checkpoint(track):
        assert neil_reel.current_track == track
        assert neil_reel.next_track == expected[track]

    neil_reel.announce_to = checkpoint

    # Start the reel and confirm the default sequential playback.
    async with trio.open_nursery() as nursery:
        neil_reel.start(nursery)
        assert neil_reel.current_track == neil_reel.tracks[0]
        assert neil_reel.next_track == neil_reel.tracks[1]

        # Play the reel.
        send_ch, receive_ch = trio.open_memory_channel(0)
        async with send_ch, receive_ch:
            nursery.start_soon(neil_reel.send_to_channel, send_ch)
            async for chunk in receive_ch:
                # Trigger the checkpoint callback for assertions.
                if chunk:
                    assert chunk


async def test_prefetch_next_track(audio_dest):
    """Start the next subprocess before it's played to start buffering."""

    async def checkpoint(spool):
        """Make sure we don't accumulate ffmpeg processes."""
        assert spool
        async with Spool('ps ax') | Spool('grep ffmpeg') as proclist:
            found_spool = False
            found_next = False
            for line in await proclist.readlines():
                if str(spool) in line:
                    found_spool = True
                if str(spool.next_track) in line:
                    found_next = True
            assert found_spool
            if spool.next_track:
                assert found_next
    assert checkpoint

    playlist = Reel([ffmpeg.read(_) for _ in [
        '/Users/zach/out000.wav', '/Users/zach/out001.wav',
        '/Users/zach/out002.wav', '/Users/zach/out003.wav',
        '/Users/zach/out004.wav', '/Users/zach/out005.wav',
    ]])
    # ]], a_announce_to=checkpoint)
    async with playlist | audio_dest() as player:
        await player.play()
        assert player


async def test_reel_manages_spool_send(neil_reel):
    """A reel can interrupt a spool and divert the stream to another track."""
    async with neil_reel as playlist:
        async with trio.open_nursery() as nursery:
            assert not playlist.current_track
            playlist.start(nursery)
            assert playlist.current_track.pid
            async with Spool('ps ax') | Spool('grep ffmp') as procs:
                found_it = False
                for line in await procs.readlines():
                    if str(playlist.current_track.pid) in line:
                        found_it = True
                assert found_it

            # Play the reel so we can exit.
            while True:
                chunk = await neil_reel.receive_some(16384)
                if not chunk:
                    break


# pylint: disable=too-many-locals
@pytest.mark.skip
async def test_reels_can_receive_from_channel(tmpdir, neil_reel):
    """A reel can receive a stream and pipe it to the current track."""
    async with trio.open_nursery() as nursery:
        file0 = tmpdir.join('0.wav')
        file1 = tmpdir.join('1.wav')
        file2 = tmpdir.join('2.wav')
        file3 = tmpdir.join('3.wav')
        file4 = tmpdir.join('4.wav')
        output_reel = Reel([
            ffmpeg.to_file(file0),
            ffmpeg.to_file(file1),
            ffmpeg.to_file(file2),
            ffmpeg.to_file(file3),
            ffmpeg.to_file(file4),
        ])
        output_reel.start(nursery)
        neil_reel.start(nursery)
        send_ch, receive_ch = trio.open_memory_channel(0)
        nursery.start_soon(output_reel.receive_from_channel, receive_ch)
        async with send_ch, output_reel:

            lock = trio.Lock()

            chunk0 = await neil_reel.receive_some(65536)
            LOG.debug(b'c0' + chunk0[:47] + chunk0[-47:])
            await send_ch.send(chunk0)

            async with lock:
                await output_reel.skip_to_next_track(close=False)
            chunk1 = await neil_reel.receive_some(65536)
            LOG.debug(b'c1' + chunk1[:47] + chunk1[-47:])
            await send_ch.send(chunk1)

            async with lock:
                await output_reel.skip_to_next_track(close=False)
            chunk2 = await neil_reel.receive_some(65536)
            LOG.debug(b'c2' + chunk2[:47] + chunk2[-47:])
            await send_ch.send(chunk2)

            async with lock:
                await output_reel.skip_to_next_track(close=False)
            chunk3 = await neil_reel.receive_some(65536)
            LOG.debug(b'c2' + chunk3[:47] + chunk3[-47:])
            await send_ch.send(chunk3)

            async with lock:
                await output_reel.skip_to_next_track(close=False)
            chunk4 = b''
            while True:
                _chunk4 = await neil_reel.receive_some(16384)
                if not _chunk4:
                    break
                chunk4 += _chunk4
                await send_ch.send(_chunk4)
            LOG.debug(b'c2' + chunk4[:47] + chunk4[-47:])

        assert await trio.Path(file0).read_bytes() == chunk0
        assert await trio.Path(file1).read_bytes() == chunk1
        assert await trio.Path(file2).read_bytes() == chunk2
        assert await trio.Path(file3).read_bytes() == chunk3
        assert await trio.Path(file4).read_bytes() == chunk4


@pytest.mark.skip
async def test_reels_can_receive_via_send_all(tmpdir, neil_reel):
    """Receive a stream incrementally in a reel with send_all."""
    # Note - check for prefetching spools with side effects, ie write files..
    async with trio.open_nursery() as nursery:
        file0 = tmpdir.join('0.wav')
        file1 = tmpdir.join('1.wav')
        file2 = tmpdir.join('2.wav')
        output_reel = Reel([
            ffmpeg.to_file(file0),
            ffmpeg.to_file(file1),
            ffmpeg.to_file(file2),
        ])
        output_reel.start(nursery)
        neil_reel.start(nursery)
        async with output_reel, neil_reel:

            # file0
            chunk0 = await neil_reel.receive_some(16384)
            await output_reel.send_all(chunk0)

            # file1
            await output_reel.skip_to_next_track()
            chunk1 = await neil_reel.receive_some(16384)
            await output_reel.send_all(chunk1)

            # file2 - flush the reel to make sure we have complete audio data
            await output_reel.skip_to_next_track()
            chunk2 = b''
            while True:
                _chunk2 = await neil_reel.receive_some(16384)
                if not _chunk2:
                    break
                chunk2 += _chunk2
                await output_reel.send_all(chunk2)

        assert await trio.Path(file0).read_bytes() == chunk0
        assert await trio.Path(file1).read_bytes() == chunk1
        file2_bytes = await trio.Path(file2).read_bytes()

        # The output file is structured differently than input data,
        # so for now it works to just compare the beginning and ending
        # bytes.  I guess..??
        assert file2_bytes[:47] == chunk2[:47]
        assert file2_bytes[-47:] == chunk2[-47:]


async def test_reel_daemon_playing(neil_reel, audio_dest):
    """Detect when a reel is playing."""
    got_here = False
    got_there = False
    async with trio.open_nursery() as nursery:
        async with neil_reel | audio_dest() as player:
            player.start_daemon(nursery)
            got_here = True
            while True:
                if player.is_done:
                    got_there = True
                    break
                await trio.sleep(0)
    assert got_here and got_there



"""The runtime configuration environment.

The config module reads settings from environment variables and
assigns defaults when appropriate.

"""
import logging
import os

from trio import Path

LOG = logging.getLogger(__name__)

__all__ = [
    'get_config',
    'get_package_dir', 'get_package_name',
    'get_xdg_home', 'get_xdg_config_dir',
    'get_config', 'get_xdg_cache_dir',
    'get_xdg_data_dir', 'get_xdg_runtime_dir',
]


async def get_package_dir():
    """Return the file path to this package's directory."""
    return (await Path(__file__).resolve()).parent


async def get_package_name():
    """Return the name of this package."""
    return (await get_package_dir()).name


async def get_xdg_home(choice=None):
    """Return the root configuration directories for this environment."""
    home_env = {
        'XDG_CONFIG_HOME': await Path('~/.config').expanduser(),
        'XDG_CACHE_HOME': await Path('~/.cache').expanduser(),
        'XDG_DATA_HOME': await Path('~/.local/share').expanduser(),
        'XDG_RUNTIME_DIR': await Path('~/.local/run').expanduser(),
    }

    # Override the defaults if already set in an environment variable.
    for key in home_env:
        home_env[key] = os.environ.get(key, home_env[key])

    if choice:
        return home_env[choice]
    return home_env


async def get_xdg_config_dir(app=None, feature=None):
    """Return a config directory for this app.

    Create the directory if it does not exist.

    """
    if app is None:
        app = await get_package_name()
    config_home = Path(await get_xdg_home('XDG_CONFIG_HOME'))
    config_dir = config_home / app
    if not await config_dir.exists():
        await config_dir.mkdir()
    if feature:
        config_dir = config_dir / feature
    if not await config_dir.exists():
        await config_dir.mkdir()
    return config_dir


async def get_config(config_path, template_file, **xconf):
    """Create a config file from a template."""
    config_file = config_path / template_file
    default_file = (await get_package_dir()) / 'templates' / template_file
    formatted = (await default_file.read_text()).format(**xconf)
    await config_file.write_text(formatted)

    return config_file


async def get_xdg_cache_dir(app=None):
    """Return a cache directory for this app.

    Create the directory if it does not exist.

    """
    if app is None:
        app = await get_package_name()
    cache_home = Path(await get_xdg_home('XDG_CACHE_HOME'))
    cache_dir = cache_home / app
    if not await cache_dir.exists():
        await cache_dir.mkdir()
    return await cache_dir.resolve()


async def get_xdg_data_dir(app=None):
    """Return a data directory for this app.

    Create the directory if it does not exist.

    """
    if app is None:
        app = await get_package_name()
    data_home = Path(await get_xdg_home('XDG_DATA_HOME'))
    data_dir = data_home / app
    if not await data_dir.exists():
        await data_dir.mkdir()
    return await data_dir.resolve()


async def get_xdg_runtime_dir(app=None):
    """Return a runtime directory for this app.

    Create the directory if it does not exist.

    """
    if app is None:
        app = await get_package_name()
    runtime_home = Path(await get_xdg_home('XDG_RUNTIME_DIR'))
    runtime_dir = runtime_home / app
    if not await runtime_dir.exists():
        await runtime_dir.mkdir()
    return await runtime_dir.resolve()



"""Magic startup file for pytest.

Imported before tests.  Loads fixtures and sets environment
variables for subprocess testing.

"""
import os
from tempfile import mkdtemp

import pytest
import trio
from trio import Path

# import reel
from reel import Reel
from reel.cmd import ffmpeg, sox


# Remove existing config vars except for testing.
for _env_key in os.environ.keys():
    if _env_key.startswith('REEL_'):
        if not _env_key.startswith('REEL_TESTS_'):
            del os.environ[_env_key]

# Create home directories for testing.
XDG = {
    'XDG_CONFIG_HOME': mkdtemp('xdg_config'),
    'XDG_CACHE_HOME': mkdtemp('xdg_cache'),
    'XDG_DATA_HOME': mkdtemp('xdg_data'),
    'XDG_RUNTIME_DIR': mkdtemp('xdg_runtime'),
}

# Set xdg environment variables.
for _key, _val in XDG.items():
    os.environ[_key] = _val

# Enable full test coverage for subprocesses.
os.environ['COVERAGE_PROCESS_START'] = 'setup.cfg'
os.environ['PYTHONPATH'] = '.'  # to find sitecustomize.py


def set_env(env):
    """Add some variables to the environment."""
    for key in env:
        os.environ[key] = env[key]


def unset_env(env):
    """Remove some variables from the environment."""
    for key in env:
        del os.environ[key]


@pytest.fixture
def neil_reel():
    """Return an audio stream of a short intro split into many files."""
    return Reel([ffmpeg.read(track) for track in [
        '/Users/zach/out000.wav', '/Users/zach/out001.wav',
        '/Users/zach/out002.wav', '/Users/zach/out003.wav',
        '/Users/zach/out004.wav', '/Users/zach/out005.wav',
    ]])


@pytest.fixture
def audio_dest():
    """Return an audio connection factory."""
    def audio_dest_fn():
        dest = os.environ.get('REEL_TESTS_AUDIO_DEST', '/dev/null')
        out = None
        if dest == 'speakers':
            out = sox.speakers()
        elif dest == 'udp':
            out = ffmpeg.to_udp('127.0.0.1', '9876')
        else:
            # Check for file output.
            out_path = Path(dest)
            if (dest == '/dev/null' or
                    trio.run(out_path.is_file) or
                    trio.run(out_path.is_dir)):
                out = ffmpeg.to_file(out_path)
        return out
    return audio_dest_fn


@pytest.fixture(params=['python -m reel.cli', 'reel'])
def cli_cmd(request):
    """."""
    return request.param


@pytest.fixture
def audio_uri():
    """Provide some handy audio file locations."""
    return {
        'RADIO': 'http://ice1.somafm.com/groovesalad-256-mp3',
        'SONG': ''.join([
            'https://archive.org/download/',
            'gd1977-05-08.shure57.stevenson.29303.flac16/',
            'gd1977-05-08d02t04.flac'
        ])
    }


@pytest.fixture(scope="session")
def env_home(tmp_path_factory):
    """Create an xdg environment with usable home directories."""
    return {
        'XDG_CONFIG_HOME': str(tmp_path_factory.mktemp('xdg_config')),
        'XDG_CACHE_HOME': str(tmp_path_factory.mktemp('xdg_cache')),
        'XDG_DATA_HOME': str(tmp_path_factory.mktemp('xdg_data')),
        'XDG_RUNTIME_DIR': str(tmp_path_factory.mktemp('xdg_runtime')),
    }


@pytest.fixture(scope="session")
def config_icecast():
    """Return variables to populate an icecast.xml config file."""
    return dict(
        location='Neptune',
        admin_email='sushi@trident.sea',
        password='hack-it-up',
        hostname='127.0.0.1',
        port='8666',
    )


@pytest.fixture(scope="function")
def audio_dir(tmp_path_factory):
    """Create a directory with different types of audio files."""
    rootdir = tmp_path_factory.mktemp('audio_dir')
    # output_wav = np.cos(2 * np.pi * 440 * np.arange(0, 5, 1 / 44100))
    # ... save the file ...
    # output_flac = Source('flac output.wav')
    return rootdir


@pytest.fixture
def music_dir(tmpdir):
    """Create a temporary directory with of music subdirectories."""
    music_dirs = [
        {
            'root1': [
                {'Some folder.shnf': ['asdf.wav']},
                {'some other folder': ['rr.txt', 'asdf.wav', 'eHello~!.doc']},
                {'subdir': [
                    {'subsubdir': [
                        {'subsubsubdir': ['asdfasdfasdfasdf.aac']},
                    ]},
                    [],
                ]},
                [],
                'rootfile.txt',
                {'yeah whatev': []},
            ],
        },
    ]

    def create_nodes(structure, parent_node=None):
        result = None
        for node in structure:
            if isinstance(node, dict):
                for subnode in node.keys():
                    if parent_node:
                        result = parent_node.join(subnode).mkdir()
                    else:
                        result = tmpdir.join(subnode).mkdir()
                    create_nodes(node[subnode], result)
            elif isinstance(node, list):
                for subnode in node:
                    create_nodes(subnode, parent_node)
            elif isinstance(node, str):
                if parent_node:
                    parent_node.join(node).write('yup')
        return result
    return str(create_nodes(music_dirs))



"""Track class."""
import logging

import trio

from ._streamer import Streamer
from ._transport import Transport

LOG = logging.getLogger(__name__)


class Track(trio.abc.AsyncResource, Streamer):
    """Something that can produce a stream."""

    def __init__(self, func):
        """Store a function for later use."""
        self._func = func
        self._nursery = None
        self._stdout = None
        self._snd_ch, self._rcv_ch = trio.open_memory_channel(0)

    def __or__(self, the_other_one):
        """Create a `Transport` out of the first two spools in the chain."""
        return Transport(self, the_other_one)

    async def __aenter__(self):
        """Run through a tranport in an async managed context."""
        return Transport(self)

    async def aclose(self):
        """Close down the Track."""

    def start(self, nursery, stdin=None):
        """Begin the process of creating a data stream."""
        self._nursery = nursery
        if stdin:
            nursery.start_soon(self.send_all, stdin)

    async def send_all(self, chunk):
        """Send a chunk of data to the input of this stream."""
        async with self._snd_ch:
            await self._snd_ch.send(self._func(chunk))
            await self._snd_ch.send(None)

    async def receive_some(self, max_bytes=65536):
        """Return a chunk of data from the output of this stream."""
        # async with self._rcv_ch:
        result = await self._rcv_ch.receive()
        if not result:
            await self._rcv_ch.aclose()
        return result



"""Tests for the keyboard module."""
from reel.keyboard import (
    Keyboard,
    KEY_LEFT, KEY_RIGHT, KEY_UP, KEY_DOWN, KEY_ESC,
    pty_stdin
)


def test_keyboard_context():
    """Open a keyboard and write some characters to it."""
    with pty_stdin('abc' * 1024) as stdin:
        with Keyboard() as keyboard:
            assert keyboard.read() == 'a'
            assert keyboard.read() == 'b'
            assert keyboard.read() == 'c'

            stdin('x')  # Goes to end of buffer

            # Drain buffer
            assert keyboard.read() == 'a'
            assert keyboard.read() == 'b'
            assert keyboard.read() == 'c'
            for _ in range(1018):
                assert keyboard.read()

            assert keyboard.read() == 'x'

            stdin('de')
            assert keyboard.read() == 'd'
            assert keyboard.read() == 'e'

            assert keyboard.read() == ''


def test_arrows():
    """Test the arrow keys."""
    k_left = bytearray([27, 91, 68])
    k_right = bytearray([27, 91, 67])
    k_down = bytearray([27, 91, 66])
    k_up = bytearray([27, 91, 65])
    k_esc = bytearray([27])
    with pty_stdin(k_left) as stdin:
        with Keyboard() as keyboard:
            assert keyboard.read() == KEY_LEFT
            stdin(k_right)
            assert keyboard.read() == KEY_RIGHT
            stdin(k_down)
            assert keyboard.read() == KEY_DOWN
            stdin(k_up)
            assert keyboard.read() == KEY_UP
            stdin(k_esc)
            assert keyboard.read() == KEY_ESC


async def test_keyboard_context_async():
    """The keyboard can be opened as a channel."""
    message = ['O', 'n', 'c', 'e', ' ', 'u', 'p', 'o']
    with pty_stdin(''.join(message)) as stdin:
        async with Keyboard() as keyboard:
            keyboard.break_on_eot('o')
            async for key in keyboard:
                assert key == message.pop(0)
            stdin('a')
            assert keyboard.read() == 'a'



# pylint: disable=W0611, W0621
"""Test the application configuration system."""
import trio
from trio import Path

from reel.config import (
    get_config,
    get_package_dir,
    get_package_name,
    get_xdg_cache_dir,
    get_xdg_config_dir,
    get_xdg_data_dir,
    get_xdg_runtime_dir,
    get_xdg_home,
)

from .conftest import set_env, unset_env


async def test_fixture_env_home_works(env_home):
    """Make sure that the env_home fixture returns usable directories."""
    for var in ['XDG_CONFIG_HOME',
                'XDG_CACHE_HOME',
                'XDG_DATA_HOME',
                'XDG_RUNTIME_DIR']:
        assert await Path(env_home[var]).exists()


async def test_fixture_env_home_uses_environ(env_home):
    """Use XDG_* environment variables when they are set."""
    set_env(env_home)
    config = await get_xdg_home()
    assert config['XDG_CONFIG_HOME'] == env_home['XDG_CONFIG_HOME']
    assert config['XDG_CACHE_HOME'] == env_home['XDG_CACHE_HOME']
    assert config['XDG_DATA_HOME'] == env_home['XDG_DATA_HOME']
    assert config['XDG_RUNTIME_DIR'] == env_home['XDG_RUNTIME_DIR']


async def test_fixture_env_home_uses_default(env_home):
    """Provide xdg home directories if not set in the environemnt."""
    unset_env(env_home)
    config = await get_xdg_home()
    home = await Path.home()
    assert str(config['XDG_CONFIG_HOME']) == str(home / '.config')
    assert str(config['XDG_CACHE_HOME']) == str(home / '.cache')
    assert str(config['XDG_DATA_HOME']) == str(home / '.local' / 'share')
    assert str(config['XDG_RUNTIME_DIR']) == str(home / '.local' / 'run')
    # ... log warning if no xdg_runtime pylint : disable = W0511 ...


async def test_get_package_dir():
    """Get the path to this package."""
    assert await Path('./reel').resolve() == await get_package_dir()


async def test_get_package_name():
    """Get the name of this package."""
    assert await get_package_name() == 'reel'


async def test_get_xdg_home(env_home):
    """Call get_xdg_home with various arguments."""
    set_env(env_home)
    config = await get_xdg_home()
    assert config['XDG_DATA_HOME'] == env_home['XDG_DATA_HOME']
    assert await get_xdg_home('XDG_DATA_HOME') == env_home['XDG_DATA_HOME']


async def test_get_xdg_config_dir(env_home):
    """Create a configuration directory for an app."""
    set_env(env_home)

    sub_dir = await get_xdg_config_dir('_testapp_', '_feature_')
    assert await sub_dir.exists()
    assert sub_dir.name == '_feature_'
    depth = len(sub_dir.parts)
    assert sub_dir.parent.name == '_testapp_'

    config_dir = await get_xdg_config_dir('_testapp_')
    assert await config_dir.exists()
    assert config_dir.name == '_testapp_'
    assert len(config_dir.parts) == depth - 1


async def test_get_xdg_config_dir_flipped(env_home):
    """Create a configuration directory for an app with the parent first."""
    set_env(env_home)

    config_dir = await get_xdg_config_dir('_testapp_')
    assert await config_dir.exists()
    assert config_dir.name == '_testapp_'
    depth = len(config_dir.parts)

    sub_dir = await get_xdg_config_dir('_testapp_', '_feature_')
    assert await sub_dir.exists()
    assert sub_dir.name == '_feature_'
    assert len(sub_dir.parts) == depth + 1
    assert sub_dir.parent.name == '_testapp_'


async def test_get_config(env_home, config_icecast):
    """Create a default config file for a command."""
    set_env(env_home)
    config_dir = await get_xdg_config_dir()
    config_file = config_dir / 'icecast.xml'
    assert not await config_file.exists()
    config = await get_config(config_dir, 'icecast.xml', **config_icecast)
    assert await config.exists()
    config_text = await config.read_text()
    for _, value in config_icecast.items():
        assert value in config_text


async def test_get_xdg_cache_dir(env_home):
    """Create a cache directory for an app."""
    set_env(env_home)

    # Check that a directory was created.
    cache_dir = await get_xdg_cache_dir('_testapp_')
    assert await cache_dir.exists()
    assert cache_dir.name == '_testapp_'

    # Make sure get_xdg_cache_dir returns a resolved path.
    xdg_cache_home = await Path(env_home['XDG_CACHE_HOME']).resolve()
    assert cache_dir.parent == xdg_cache_home


async def test_get_xdg_data_dir(env_home):
    """Create a data directory for an app."""
    set_env(env_home)

    # Check that a directory was created.
    data_dir = await get_xdg_data_dir('_testapp_')
    assert await data_dir.exists()
    assert data_dir.name == '_testapp_'

    # Make sure get_xdg_data_dir returns a resolved path.
    xdg_data_home = await Path(env_home['XDG_DATA_HOME']).resolve()
    assert data_dir.parent == xdg_data_home


async def test_get_xdg_runtime_dir(env_home):
    """Create a runtime directory for an app."""
    set_env(env_home)

    # Check that a directory was created.
    runtime_dir = await get_xdg_runtime_dir('_testapp_')
    assert await runtime_dir.exists()
    assert runtime_dir.name == '_testapp_'

    # Make sure get_xdg_runtime_dir returns a resolved path.
    xdg_runtime_home = await Path(env_home['XDG_RUNTIME_DIR']).resolve()
    assert runtime_dir.parent == xdg_runtime_home



"""Tests for :mod:`reel.proc`."""
from trio import move_on_after, Path

from reel import __version__, Spool


async def test_class_spool(capsys):
    """Run a process and read stdout."""
    src = Spool('python   -m   reel.cli  -v')  # ... make separate tests
    version = await src.run()
    assert src.returncode == 0
    assert src.stderr is None
    assert version == __version__

    # not sure why here..  from another test, but test logging somwhere?
    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''


async def test_spool_simple_command():
    """Run a simple command with :class:`reel.Spool`."""

    # Use :func:`Spool.run` to read the output of a command.
    china = await Spool('cat').run('sunflower')
    assert china == 'sunflower'

    # Use :class:`reel.Transport` to read the output in a managed context.
    async with Spool('cat') as china:
        assert await china.read('sunflower') == 'sunflower'


async def test_pwd():
    """Test a familiar shell command."""
    assert str(await Path.cwd()) == await Spool('pwd').run()


async def test_limit_output():
    """Set a cutoff limit on output."""
    infinity = Spool('cat /dev/zero').limit(47)
    order = len(await infinity.run())
    assert order == 47


async def test_source_run_timeout():
    """Make sure a process returns normally with a timeout interrupt."""
    random = Spool('cat /dev/random')
    with move_on_after(0.47):
        await random.run()
    assert random.stdout and random.stdout != b''
    # ??? cant get return value from run() on timeout
    # ??? is the process still running when we wait?
    # await random._proc.wait()
    # assert random.returncode <= 0


async def test_stderr():
    """Run a process and read stderr."""
    not_here = Spool('ls /not_here_i_hope')
    assert not await not_here.run()  # no output
    assert not_here.stderr  # some error about a missing file
    assert not_here.returncode  # error present
    assert 'not_here' in not_here.stderr



"""Reel class."""
import logging

import trio

from ._streamer import Streamer
from ._transport import Transport

LOG = logging.getLogger(__name__)


class Reel(trio.abc.AsyncResource, Streamer):
    """A stack of spools concatenated in place as one spool in a transport."""

    def __init__(self, tracks, announce_to=None, a_announce_to=None):
        """Begin as a list of tracks."""
        self._a_announce = a_announce_to
        self._announce = announce_to
        self._current_track = None
        self._next_track = None
        self._nursery = None
        self._stdin = None
        self._tracks = tracks

    def __str__(self):
        """Print the command."""
        return self.__repr__()

    def __repr__(self):
        """Represent prettily."""
        result = 'Reel([ '
        for track in self._tracks:
            result += track.__repr__()
            result += ','
        result += ' ])'
        return result

    def __or__(self, next_one):
        """Combine with the next one as a transport."""
        return Transport(self, next_one)

    def __rshift__(self, next_one):
        """Combine with the next one as a transport."""
        return Transport(self, next_one)

    async def aclose(self):
        """Close the spools."""
        for track in self._tracks:
            await track.aclose()

    @property
    def announce_to(self):
        """Return the next track on deck."""
        return self._announce

    @announce_to.setter
    def announce_to(self, value):
        """Set the announce_to callback."""
        self._announce = value

    @property
    def announce_to_async(self):
        """Return the next track on deck."""
        return self._a_announce

    @announce_to_async.setter
    def announce_to_async(self, value):
        """Set the announce_to callback."""
        self._a_announce = value

    @property
    def next_track(self):
        """Return the next track on deck."""
        if self._next_track:
            return self.tracks[self._next_track]
        return None

    @property
    def current_track(self):
        """Return the currently playing track."""
        if self._current_track is not None:
            return self._tracks[self._current_track]
        return None

    @property
    def tracks(self):
        """Return the list of tracks."""
        return self._tracks

    @property
    def spools(self):
        """Return the list of spools."""
        return self._tracks

    def start(self, nursery, stdin=None):
        """Store information for send to use."""
        LOG.debug('[ START REEL %d ]', len(self._tracks))
        self._nursery = nursery
        self._stdin = stdin
        self._start_next_track()

    def _start_next_track(self):
        """Set the current/next track so send has something to send."""
        LOG.debug(
            '[ REEL %d START NEXT_TRACK %s %s ]',
            len(self._tracks),
            str(self._current_track),
            str(self._next_track)
        )

        # Start at the beginning without a current track.
        if not self.current_track:
            if self._tracks:
                self._current_track = 0
                if len(self._tracks) > 1:
                    self._next_track = 1
            self.current_track.start(self._nursery, self._stdin)
        else:
            _next = None
            if self._next_track:
                _next = self._next_track + 1
            self._current_track = self._next_track
            if _next and _next < len(self._tracks):
                self._next_track = _next
            else:
                self._next_track = None

        # Prefetch next track by starting it.
        if self.next_track:
            self.next_track.start(self._nursery, self._stdin)

        LOG.debug(
            '[ REEL %d STARTED NEXT_TRACK %s %s ]',
            len(self._tracks),
            str(self._current_track),
            str(self._next_track)
        )

    async def skip_to_next_track(self, close=True):
        """Begin playing the next track immediately."""
        LOG.debug(
            '[ REEL %d SKIP_TO_NEXT_TRACK %s %s ]',
            len(self._tracks),
            self._current_track,
            self._next_track
        )
        if close:
            await self.current_track.aclose()

        if self.next_track:
            self._start_next_track()

            # Announce the track change.
            if self._announce:
                self._announce(self.current_track)
            elif self._a_announce:
                await self._a_announce(self.current_track)

    async def receive_from_channel(self, channel):
        """Receive input and send it to the current spool."""
        LOG.debug(
            '[ REEL track[%s] receive_from_channel() ]',
            str(self._current_track)
        )
        async with channel:
            async for chunk in channel:
                await self.send_all(chunk)

    async def send_all(self, chunk):
        """Receive input and send it to the current spool."""
        LOG.debug(
            '[ REEL track[%s] send_all !! len %d ]',
            str(self._current_track),
            len(chunk)
        )
        if self.current_track:  # race??  next line could be error?
            await self.current_track.send_all(chunk)

    async def receive_some(self, max_bytes):
        """Return a chunk of data from the output of this stream."""
        if self.current_track:

            # Return a chunk of data from the current track.
            chunk = await self.current_track.receive_some(max_bytes)
            if chunk:
                return chunk

            # No data, close the track and start the next one.
            await self.current_track.proc.aclose()
            if self.next_track:
                self._start_next_track()

                # Announce the track change.
                if self._announce:
                    self._announce(self.current_track)
                elif self._a_announce:
                    await self._a_announce(self.current_track)

                # Return a chunk of data from the new track.
                return await self.receive_some(max_bytes)

        # Send empty byte as EOF.
        return b''

    async def send_to_channel(self, channel):
        """Start each spool and send stdout to the `channel`."""
        async with channel:
            while self.current_track:

                # Announce the track change.
                if self._announce:
                    self._announce(self.current_track)
                elif self._a_announce:
                    await self._a_announce(self.current_track)

                # Play the track.
                while True:
                    chunk = await self.receive_some(16384)
                    if chunk:
                        await channel.send(chunk)
                    else:
                        break
                self._start_next_track()



"""Streamer class."""
import abc
import logging

LOG = logging.getLogger(__name__)


class Streamer(metaclass=abc.ABCMeta):
    """Something that can stream i/o in a transport."""

    @abc.abstractmethod
    def start(self, nursery, stdin=None):
        """Start whatever this thing needs to run.

        Called before send or receive.

        """

    async def stop(self):
        """Stop it."""

    async def send_to_channel(self, channel):
        """Send data to the channel and close both sides."""
        LOG.debug('-Streamer- send_to_channel')
        async with channel:
            while True:
                chunk = await self.receive_some(65536)
                if chunk:
                    await channel.send(chunk)
                else:
                    break

    async def receive_from_channel(self, channel):
        """Receive data from the channel."""
        async with channel:
            async for chunk in channel:
                await self.send_all(chunk)

    @abc.abstractmethod
    async def receive_some(self, max_bytes):
        """Return a chunk of data from this stream's output."""

    @abc.abstractmethod
    async def send_all(self, chunk):
        """Send a chunk of data to this stream's input."""



"""Command line interface to :mod:`reel`."""
import argparse
import logging
import os
import sys

import trio

import reel

if 'REEL_LOG_LEVEL' in os.environ:
    LOG_DIR = trio.run(reel.config.get_xdg_data_dir, 'reel')
    LOG_FILE = LOG_DIR / 'cli.log'
    LOG_LEVEL = os.environ.get('REEL_LOG_LEVEL').upper()
    logging.basicConfig(filename=LOG_FILE, level=LOG_LEVEL)
LOG = logging.getLogger(__name__)
LOG.debug('Begin logging for reel <~-~-~(~<~(o~>)~)~-~-~>')


async def main() -> int:
    """Test."""
    await trio.sleep(0)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c', '--config', action='store_true',
        help='print the configuration'
    )
    parser.add_argument(
        '-v', '--version', action='store_true',
        help='print the version'
    )
    args = parser.parse_args()

    if args.config:
        LOG.debug('args.config: %s', args.config)
        if 'REEL_LOG_LEVEL' in os.environ:
            print(f'REEL_LOG_DIR={LOG_DIR}')
            print(f'REEL_LOG_FILE={LOG_FILE}')
            print(f'REEL_LOG_LEVEL={LOG_LEVEL}')

    elif args.version:
        LOG.debug('args.version: %s', args.version)
        print(reel.__version__)

    else:
        LOG.debug('no args')
        print('leer')

    return 0


def enter():
    """Run the main program as an ``entry_point`` for :mod:`setuptools`."""
    sys.exit(trio.run(main))


if __name__ == '__main__':
    enter()



"""Test the API."""
import logging

import trio
from trio import Path

from reel import Daemon, Server, Spool
from reel.config import get_config, get_xdg_config_dir

logging.basicConfig(
    filename='/Users/zach/asdf.log',
    level='DEBUG',
    format='%(levelname)s:%(lineno)d:%(module)s:%(funcName)s:%(message)s'
)
LOG = logging.getLogger(__name__)


async def test_shell_commands():
    """Run a few shell commands."""
    assert 'rutabaga' in await Spool(f'grep utabag {__file__}').run()

    async with Spool('find .', xflags=['-type', 'f']) as found:
        lines = await found.readlines()
        found_resolved = [str(await Path(_).resolve()) for _ in lines]
        assert __file__ in found_resolved

    assert float((await Spool('python -V').run())[7:10]) >= 3.5

    xconf = ['-c', "import os; print(os.environ['_ASDF'])"]
    env = {'_ASDF': 'asdf'}
    assert await Spool('python', xflags=xconf, xenv=env).run() == 'asdf'

    # fname = 'test.flac'
    # assert await Source(f'flac -t {fname}').read_bool()
    # assert isinstance(await Source(f'flac -t {fname}').read_bool(), bool)

    # async with Source(f'flac -t {fname}') as test:
    #     assert await test.read_bool()
    #     assert test.status == 0
    #     assert 'test.flac: ok' in test.err

    # wav = await Source(f'flac -cd {fname}').read_bytes()
    # assert len(wav) == 57684956
    # flac = await Source(f'flac -c -').read_bytes(send_bytes=wav)
    # assert len(flac) == 27583002


async def test_server(config_icecast):
    """Run a server."""
    config_dir = await get_xdg_config_dir()
    config = await get_config(config_dir, 'icecast.xml', **config_icecast)
    flags = ['-c', str(config)]

    # Start an icecast daemon.
    LOG.debug('------>><<>><< before open_nursery')
    async with trio.open_nursery() as nursery:
        LOG.debug('------>><<>><< before async with daemon')
        async with Daemon('icecast', xflags=flags) > nursery as icecast:
            LOG.debug('------>><<>><< after async with daemon')
            assert isinstance(icecast, Server)
            # nursery.start_soon(icecast.run, nursery)
            # icecast.start_daemon(nursery)

            await trio.sleep(0.5)  # give icecast time to start

            # Make sure it started.
            async with Spool('ps ax') | Spool('grep icecast') as procs:
                LOG.debug('------>><<>><< after async with spool')
                found_it = False
                lines = await procs.readlines()
                LOG.debug('------>><<>><< BEFORE lines')
                for line in lines:
                    LOG.debug('------>><<>><< IN lines')
                    if str(config) in line:
                        found_it = True
                # ... WEIRD!! 'assert False' here hangs!!!!!
                assert found_it
            # await icecast.stop()
            LOG.debug('------>><<>><< OUT of spool')
        LOG.debug('------>><<>><< OUT of daemon')
    LOG.debug('------>><<>><< OUT of nursery')

    # Check the process list for no zombies.
    async with Spool('ps ax') as proclist:
        found = False
        for line in await proclist.readlines():
            if str(config) in line:
                found = True
        assert not found



"""The background process."""
import logging
import subprocess

import trio

from reel.config import get_config, get_xdg_config_dir
from ._server import Server
from ._spool import Spool

LOG = logging.getLogger(__name__)


class Daemon(Spool):
    """A background process."""

    _config_base = None
    _config = dict()

    async def _prepare(self, config):
        """Prepare the config."""

    def __or__(self, next_one):
        """Create a server with the first two spools."""
        return Server([self, next_one])

    def __gt__(self, nursery):
        """Set the nursery to use in the context manager."""
        return Server(self) > nursery

    def __init__(self, command=None, xenv=None, xflags=None):
        """Init the :class:`~reel.Spool`."""
        if command:
            self._command = command
        super().__init__(self._command, xenv=xenv, xflags=xflags)

    async def aclose(self):
        """Terminate the process."""
        if self._proc:
            self._proc.terminate()

    async def __aenter__(self):
        """Open a server context for this daemon."""
        return Server(self)

    async def launch(self, nursery, task_status=trio.TASK_STATUS_IGNORED):
        """Run the daemon."""
        if self._config_base:
            config = await get_config(
                await get_xdg_config_dir(),
                self._config_base,
                **self._config
            )

            # Give the subclasses a chance to fill in configuration vars
            await self._prepare(config)

        self._proc = trio.Process(
            self._command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self._env
        )
        nursery.start_soon(self._handle_stdout)
        nursery.start_soon(self._handle_stderr)
        task_status.started()



"""Pre-configured commands."""
from . import ffmpeg, icecast, sox

from ._aria2 import Aria2
from .icecast import Icecast
from ._redis import Redis

SRC_SILENCE = "ffmpeg -re -f s16le -i /dev/zero -f s16le -"

SRC_FILE = "ffmpeg -ac 2 -i {filename} -f s16le -ar 44.1k -acodec pcm_s16le -"


PYP_NORMED = """fmpeg -ac 2 -i {filename} -af loudnorm=I=-16:TP=-1.5:LRA=11
                -ac 2 -f s16le -ar 44.1k -acodec pcm_s16le -"""


DST_UDP = """ffmpeg -re -ac 2 -ar 44.1k -f s16le -i -
             -vn -acodec mp3 -q:a 0 -f mp3 udp://{ipaddress}:{port}"""

DST_ICECAST = """ffmpeg -re -ac 2 -ar 44.1k -f s16le -i -
                 -vn -acodec mp3 -q:a 0 -f mp3
                 icecast://source:{password}@{host}:{port}/{mount}"""

DST_SPEAKER = """play -t raw -r 44.1k -e signed-integer
                 -b 16 --endian little -c 2 -"""



"""Test the pipes."""
# pylint: disable=W0611, W0613, W0621
import logging

import reel


async def test_pipe_operator():
    """Overload the ``__or__`` operator to make piping streams look cool."""
    logging.debug('+++++++++++++++++++++>> BEGIN TEST_PIPE_OPERATOR')

    # Spool some commands
    read_file = reel.Spool(f'cat {__file__}')
    remove_grep = reel.Spool(f'grep -v grep')
    find_cat = reel.Spool('grep cat')

    # One way to make a transport
    transport = reel.Transport(read_file, remove_grep, find_cat)

    # Another way
    chain = [read_file, remove_grep, find_cat]
    transport_chain = reel.Transport(chain)

    # Or:
    async with read_file | remove_grep | find_cat as out:
        assert repr(transport) == repr(transport_chain) == repr(out)
        lines = await out.readlines()
        assert len(lines) == 2
        for line in lines:
            assert 'cat' in line


async def test_play_music_even_better(audio_dest):  # noqa: F811
    """Try pipe operators.
    logging.debug('+++++++++++++++++++++>> BEGIN TEST_PLAY_MUSIC_EVEN_BETTER')

    This method is not gapless.  Reconnecting to the speaker
    on each iteration can (will?) definitely produce some noise.
    Looks cool, though.

    """
    logging.debug('////////////// gaps //////////////////')
    for part in [f'/Users/zach/out00{_}.wav' for _ in range(0, 6)]:
        src = reel.cmd.ffmpeg.read(part).limit(1024**2)
        async with src | audio_dest() as player:
            assert isinstance(player, reel.Transport)
            await player.play()


async def test_play_neil_with_pipes(audio_dest, audio_uri):  # noqa: F811
    """Try pipe operators on the *Come on Baby Let's Go Downtown* intro.

    This method is gapless.  The `Transport` can keep the speaker
    open while switching files.

    """
    logging.debug('+++++++++++++++++++++>> BEGIN TEST_PLAY_NEIL_WITH_PIPES')
    logging.debug('////////////// nogaps ////////////////')
    playlist = reel.Reel([reel.cmd.ffmpeg.read(_).limit(1024**2) for _ in [
        '/Users/zach/out000.wav', '/Users/zach/out001.wav',
        '/Users/zach/out002.wav', '/Users/zach/out003.wav',
        '/Users/zach/out004.wav', '/Users/zach/out005.wav',
        # audio_uri['SONG'], audio_uri['RADIO'],
    ]])
    async with playlist | audio_dest() as player:
        await player.play()



"""This is reel."""
import pkg_resources as _pkg
__version__ = _pkg.get_distribution(__name__).version

from . import cmd
from . import config
from ._daemon import Daemon
from ._reel import Reel
from ._server import Server
from ._spool import Spool
from ._streamer import Streamer
from ._track import Track
from ._transport import Transport



"""Test suite for the :mod:`reel.cli` command line interface."""
import reel


async def test_no_args(cli_cmd):
    """Run :mod:`reel.cli` with no arguments."""
    noargs = reel.Spool(cli_cmd)
    output = await noargs.run()
    assert noargs.returncode == 0
    assert output == 'leer'


async def test_version(cli_cmd):
    """Show the version number."""
    for flag in ['-v', '--version']:
        command = f'{cli_cmd} {flag}'
        version = reel.Spool(command)
        output = await version.run()
        assert version.returncode == 0
        assert output == reel.__version__


async def test_config(cli_cmd):
    """Show the current configuration."""
    for flag in ['-c', '--config']:
        command = f'{cli_cmd} {flag}'
        config = reel.Spool(command, {'REEL_LOG_LEVEL': 'debug'})
        dump = (await config.run()).split('\n')
        assert len(dump) == 3
        assert config.returncode == 0
        for line in dump:
            assert line.startswith('REEL_')



"""Use the sox play utility."""
from .._spool import Spool


def speakers():
    """Send audio to the speaker."""
    cmd = 'play'
    flags = ['-t', 'raw',
             '-r', '44.1k',
             '-e', 'signed-integer',
             '-b', '16',
             '--endian', 'little',
             '-c', '2',
             '-']
    return Spool(cmd, xflags=flags)



"""Test the pre-configured commands."""
# import trio

import reel

BYTE_LIMIT = 1000000


async def test_audio_dir(audio_dir):
    """Get test audio files."""
    assert audio_dir.exists()


async def test_import():
    """Make sure the module is imported."""
    assert reel.cmd
    assert reel.cmd.SRC_SILENCE


# async def test_icecast():
#     """Run an icecast server."""
#     cmd = reel.cmd.icecast.Icecast('127.0.0.1', '8676', 'mnt', 'pw')
#     async with cmd as icecast:
#         await icecast.start_daemon()
#         await trio.sleep(1)
#         await icecast.stop()
#
#     async with trio.open_nursery() as nursery:
#         daemon = await reel.Transport(ice).start_daemon(nursery)
#         assert daemon
#         trio.sleep(1)
#         await daemon.stop()
#         assert daemon.returncode == 0


# async def test_icecast():
#     """Run an icecast server."""
#     ice = reel.cmd.icecast.Icecast('127.0.0.1', '8676', 'mnt', 'pw')
#     async with trio.open_nursery() as nursery:
#         daemon = await reel.Transport(ice).start_daemon(nursery)
#         assert daemon
#         trio.sleep(1)
#         await daemon.stop()
#         assert daemon.returncode == 0



"""Install the reel package."""
from setuptools import find_packages, setup

setup(
    license='MIT',
    name='reel',
    version=open('VERSION', 'r').read().strip(),
    author='Zach Thompson',
    author_email='zach@allotropic.com',
    url='http://github.com/zthompson47/reel',
    description='A contraption for async subprocess streaming',
    long_description_content_type='text/x-rst',
    long_description=open('README.rst', 'r').read(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts': ['reel=reel.cli:enter']},
    install_requires=['trio>=0.10.0'],
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
)



"""Various ffmpeg command line tools."""
from .._spool import Spool


def read(uri):
    """Prepare a command to read an audio file and stream to stdout."""
    cmd = 'ffmpeg'
    flags = [
        '-ac', '2',  # 2-channel stereo
        '-i', uri,  # input file or url
        '-f', 's16le',  # 16 bit little-endian
        '-ar', '44.1k',  # sample rate
        '-acodec', 'pcm_s16le',  # wav format
        '-',  # stream to stdout
    ]
    return Spool(cmd, xflags=flags)


def to_icecast(host, port, mount, password):
    """Stream audio to an icecast server."""
    cmd = 'ffmpeg'
    flags = [
        '-re',
        '-ac', '2',
        '-ar', '44.1k',
        '-f', 's16le',
        '-i', '-',
        '-vn',
        '-codec:a', 'libvorbis',
        '-q:a', '8.0',
        '-content_type', 'audio/ogg',
        '-f', 'ogg',
        f'icecast://source:{password}@{host}:{port}/{mount}'
    ]
    return Spool(cmd, xflags=flags)


def to_udp(host, port):
    """Stream audio over udp."""
    cmd = 'ffmpeg'
    flags = [
        '-re',  # realtime flow control
        '-ac', '2',  # 2-channel stereo
        '-ar', '44.1k',  # sample rate
        '-f', 's16le',  # 16 bit littl-endian
        '-i', '-',  # receive from stdin
        '-vn',  # no video
        '-acodec', 'mp3',  # compress to mp3
        '-q:a', '0',  # maximum quality
        '-f', 'mp3',  # mp3 format
        f'udp://{host}:{port}',  # receiver address
    ]
    return Spool(cmd, xflags=flags)


def to_file(path):
    """Stream to a file."""
    cmd = 'ffmpeg'
    flags = [
        '-ac', '2',  # 2-channel stereo
        '-ar', '44.1k',  # sample rate
        '-f', 's16le',  # 16 bit little-endian
        '-i', '-',  # receive from stdin
        '-f', 's16le',  # 16 bit little-endian
        '-ar', '44.1k',  # sample rate
        '-y',  # overwrite output file
        '-c:a', 'copy',  # stream copy audio
        path,
    ]
    return Spool(cmd, xflags=flags)



"""A keylogger context manager."""
from contextlib import contextmanager
import fcntl
import io
import logging
import os
import pty
import sys
import termios
import time
import tty

import trio

LOG = logging.getLogger(__name__)

KEY_EOT = b'\x04'
KEY_ESC = b'\x1b'

KEY_UP = b'\x1b[A'
KEY_DOWN = b'\x1b[B'
KEY_RIGHT = b'\x1b[C'
KEY_LEFT = b'\x1b[D'


class Keyboard:
    """A way to get keyboard input on Posix systems."""

    def __init__(self):
        """Store the term settings before going into cbreak mode."""
        self._break_on_eot = None
        self._stashed_term = None

        # Set stdin file descriptor to non-blocking mode
        fileno = sys.stdin.fileno()
        flag = fcntl.fcntl(fileno, fcntl.F_GETFL)
        fcntl.fcntl(fileno, fcntl.F_SETFL, flag | os.O_NONBLOCK)

    async def __aenter__(self):
        """Alias the sync func."""
        return self.__enter__()

    def __enter__(self):
        """Save the current term and create a new one with no echo."""
        try:
            self._stashed_term = termios.tcgetattr(sys.stdin)
            tty.setcbreak(sys.stdin, termios.TCSANOW)
        except termios.error:
            LOG.debug('No termios', exc_info=True)
        return self

    async def __aexit__(self, *args):
        """Bring back the original term settings."""
        self.__exit__(args)

    def __exit__(self, *args):
        """Bring back the original term settings."""
        if self._stashed_term:
            termios.tcsetattr(
                sys.stdin,
                termios.TCSANOW,
                self._stashed_term
            )

    def __aiter__(self):
        """Return iterator object."""
        return self

    async def __anext__(self):
        """Iterate over the input streamReturn the next character."""
        with block_stdin(True):
            char = await trio.run_sync_in_worker_thread(
                self.read,
                cancellable=True
            )
        if char == self._break_on_eot:
            raise StopAsyncIteration
        return char

    def break_on_eot(self, character=KEY_EOT):
        """Set an eot character to break from an iterator."""
        self._break_on_eot = character

    @staticmethod
    def read():
        """Read a unicode character from stdin.

        https://en.wikipedia.org/wiki/ANSI_escape_code#Escape_sequences
        """
        char = sys.stdin.read(1)
        if char != '' and ord(char) == ord(KEY_ESC):

            with block_stdin(False):
                csi = sys.stdin.read(1)
                if csi == '[':
                    ctl = sys.stdin.read(1)
                    if ctl == 'A':
                        return KEY_UP
                    if ctl == 'B':
                        return KEY_DOWN
                    if ctl == 'C':
                        return KEY_RIGHT
                    if ctl == 'D':
                        return KEY_LEFT

            return KEY_ESC
        return char


@contextmanager
def block_stdin(block: bool) -> None:
    """Toggle stdin blocking mode."""
    fileno = sys.stdin.fileno()
    flag = fcntl.fcntl(fileno, fcntl.F_GETFL)

    if block:
        fcntl.fcntl(fileno, fcntl.F_SETFL, flag | ~os.O_NONBLOCK)
    else:
        fcntl.fcntl(fileno, fcntl.F_SETFL, flag | os.O_NONBLOCK)

    yield

    fcntl.fcntl(fileno, fcntl.F_SETFL, flag)


@contextmanager
def pty_stdin(preload=None):
    """Replace stdin with a pty.

    Returns:
        A function to write data into stdin.

    Examples:
        >>> with pty_stdin('ab') as stdin:
        ...     with Keyboard() as keyboard:
        ...         assert keyboard.read() == 'a'
        ...         assert keyboard.read() == 'b'
        ...         stdin('x')
        ...         assert keyboard.read() == 'x'
        1

    """
    _stashed_stdin = sys.stdin
    master, slave = pty.openpty()
    sys.stdin = io.TextIOWrapper(open(slave, 'rb'))

    def stdin(text):
        """Write data to the master side of the pty.

        Returns:
            Number of bytes written.

        Examples:
            >>> with pty_stdin() as stdin:
            ...   assert(stdin('a') == 1)

        """
        if isinstance(text, str):
            return os.write(master, str.encode(text))
        return os.write(master, text)

    if preload:
        bytes_written = stdin(preload)
        LOG.debug('bytes: %d', bytes_written)

    yield stdin

    sys.stdin = _stashed_stdin


def main():
    """Test the keyboard with a polling loop."""
    with Keyboard() as keyboard:
        while True:
            time.sleep(0.01)
            char = keyboard.read()
            if char == 'q':
                break
            print(char, end='')
            sys.stdout.flush()

    print('\n')
    print('Exiting in a few seconds, term should be back to normal...')
    time.sleep(3)


async def amain():
    """Test the keyboard with a polling loop, async."""
    async with Keyboard() as keyboard:
        keyboard.break_on_eot('q')
        async for key in keyboard:
            print(key, end='')
            sys.stdout.flush()
        print('')

if __name__ == '__main__':
    # main()
    trio.run(amain)



"""Spool class."""
import logging
import os
import shlex
import subprocess

import trio

from ._transport import Transport

LOG = logging.getLogger(__name__)


class Spool(trio.abc.AsyncResource):
    """A shell command."""

    def __init__(self, command, xenv=None, xflags=None):
        """Queue a subprocess."""
        if isinstance(command, list):
            self._command = command
        else:
            self._command = shlex.split(command)
        self._env = os.environ.copy()
        self._limit = None
        self._proc = None
        self._status = None
        self._stderr = None
        self._stdout = None
        if xflags:
            for flag in xflags:
                # Accept objects like Path that look like a str
                self._command.append(str(flag))
        if xenv:
            for key, val in xenv.items():
                self._env[key] = val
        # super().__init__()
        LOG.debug(self.__repr__())

    def __str__(self):
        """Print the command."""
        return ' '.join(self._command)

    def __repr__(self):
        """Represent the command."""
        return f"Spool('{str(self)}')"

    def __or__(self, next_one):
        """Create a server with the first two spools."""
        return Transport(self, next_one)

    def __rshift__(self, next_one):
        """Create a transport with the first two spools."""
        return Transport(self, next_one)

    async def __aenter__(self):
        """Start an async context."""
        return Transport(self)

    async def aclose(self):
        """Clean up."""
        if self.proc:
            await self.proc.aclose()

    @property
    def pid(self):
        """Return the process pid."""
        if self._proc:
            return self._proc.pid
        return None

    @property
    def proc(self):
        """Return the process."""
        return self._proc

    @property
    def returncode(self):
        """Return the exit code of the process."""
        return self._proc.returncode

    @property
    def stderr(self):
        """Return whatever the process sent to stderr."""
        if self._stderr:
            return self._stderr.decode('utf-8')
        return None

    @property
    def stdout(self):
        """Return whatever the process sent to stdout."""
        # if self._stdout:
        #    return self._stdout.decode('utf-8')
        # return None
        return self._stdout

    def limit(self, byte_limit=65536):
        """Configure this `spool` to limit output to `byte_limit` bytes."""
        self._limit = byte_limit
        return self

    async def run(self, message=b''):
        """Send stdin to process and return stdout."""
        async with trio.open_nursery() as nursery:
            self._proc = trio.Process(
                self._command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=self._env
            )
            nursery.start_soon(self._handle_stdin, message)
            nursery.start_soon(self._handle_stdout, self._limit)
            nursery.start_soon(self._handle_stderr)
        await self._proc.wait()
        if self._stdout:
            return self._stdout.decode('utf-8', errors='ignore').strip()
        return None

    async def _handle_stderr(self):
        """Read stderr."""
        while True:
            try:
                chunk = await self._proc.stderr.receive_some(16384)
            except trio.ClosedResourceError:
                LOG.debug('Stderr closed', exc_info=True)
                break
            else:
                if not chunk:
                    break
                if not self._stderr:
                    self._stderr = b''
                self._stderr += chunk

    async def _handle_stdout(self, limit=None):
        """Read stdout."""
        max_rcv = limit or 16834

        while True:
            try:
                chunk = await self._proc.stdout.receive_some(max_rcv)
                LOG.debug(chunk)
            except trio.ClosedResourceError:
                LOG.debug('Stdout closed', exc_info=True)
                break
            else:
                if not chunk:
                    break
                if not self._stdout:
                    self._stdout = b''
                self._stdout += chunk
                max_rcv -= len(chunk)
                if max_rcv == 0:
                    await self._proc.stderr.aclose()
                    await self._proc.stdout.aclose()
                    break

    async def _handle_stdin(self, message):
        """Receive stdin."""
        if isinstance(message, str):
            _msg = message.encode('utf-8')
        elif message is None:
            _msg = b''
        else:
            _msg = message
        async with self.proc.stdin as stdin:
            await stdin.send_all(_msg)

    def handle_stderr(self, nursery):
        """Read stderr, called as a task in a nursery."""
        nursery.start_soon(self._handle_stderr)

    def handle_stdout(self, nursery):
        """Read stdout, called as a task in a nursery."""
        nursery.start_soon(self._handle_stdout)

    def handle_stdin(self, nursery, message):
        """Feed stdin, called as a task in a nursery."""
        nursery.start_soon(self._handle_stdin, message)

    def start(self, nursery, stdin=None):
        """Initialize the subprocess and run the command."""
        LOG.debug('-- << SPOOL start about to run proc %s', self)
        self._proc = trio.Process(
            self._command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self._env
        )
        LOG.debug('-- >> SPOOL start ljjjj to run proc %s', self._proc)
        if stdin:
            self.handle_stdin(nursery, stdin)
        self.handle_stderr(nursery)

    async def stop(self):
        """Stop it."""
        await self.aclose()

    async def receive_from_channel(self, channel):
        """Send output of channel to stdin."""
        try:
            async with self._proc.stdin:
                async for chunk in channel:
                    await self._proc.stdin.send_all(chunk)
        except trio.ClosedResourceError as error:
            LOG.debug(error)
        except AttributeError:
            LOG.debug('<><><><><><><><><><><: %s %s', self, self._proc)

    async def send_to_channel(self, channel):
        """Stream stdout to `channel` and close both sides."""
        async with channel:
            async with self.proc:
                await self.send_no_close(channel)

    async def send_all(self, chunk):
        """Send a chunk of data to stdin."""
        await self._proc.stdin.send_all(chunk)

    async def receive_some(self, max_bytes):
        """Return a chunk of data from the output of this stream."""
        try:
            return await self._proc.stdout.receive_some(max_bytes)
        except trio.ClosedResourceError as error:
            LOG.debug(error)

    async def send_no_close(self, channel):
        """Stream stdout to `channel` without closing either side."""
        buffsize = 16384
        bytes_received = 0

        # Don't receive more than the bytes limit.
        if self._limit and self._limit < 16384:
            buffsize = self._limit

        # <=~ Receive data.
        chunk = await self.receive_some(buffsize)
        while chunk:

            # ~=> Send data.
            await channel.send(chunk)

            # Check for byte limit.
            bytes_received += len(chunk)
            if self._limit and bytes_received > self._limit:
                break

            # Cap buffer at limit.
            buffsize = 16384
            if self._limit and self._limit < 16384:
                buffsize = self._limit

            # <=~ Receive data.
            chunk = await self.receive_some(buffsize)



"""Various ffprobe command line tools."""
import reel


class Devices(reel.Spool):
    """Provides information about system audio devices."""

    cmd = 'ffprobe'
    flags = ['-hide_banner', '-devices']

    def __init__(self):
        """."""
        super().__init__(self.cmd, xflags=self.flags)

    def list_outputs(self):
        """."""

    def list_inputs(self):
        """."""


class Codecs(reel.Spool):
    """Provides information about supported audio codecs."""

    cmd = 'ffprobe'
    flags = ['-hide_banner', '-codecs']

    def __init__(self):
        """."""
        super().__init__(self.cmd, xflags=self.flags)

    def audio_codecs(self):
        """."""



"""The background process supervisor."""
import logging

import trio

LOG = logging.getLogger(__name__)


class Server(trio.abc.AsyncResource):
    """A device for running daemon services."""

    def __init__(self, daemon):
        """Get list of daemons ready."""
        self._nursery = None
        if isinstance(daemon, list):
            self._daemons = daemon
        else:
            self._daemons = [daemon]

    def __gt__(self, nursery):
        """Set the nursery to use in the context manager."""
        self._nursery = nursery
        return self

    def __or__(self, next_one):
        """Supervise multiple daemon processes with bitwise or."""
        self._daemons.append(next_one)
        return self

    async def aclose(self):
        """Clean up."""
        for daemon in self._daemons:
            await daemon.aclose()

    async def __aenter__(self):
        """Start up."""
        for daemon in self._daemons:
            await self._nursery.start(daemon.launch, self._nursery)
        return self



"""The icecast streaming server."""
from .._daemon import Daemon
from .._spool import Spool


class Icecast(Daemon):
    """An icecast client and server."""

    _command = 'icecast'
    _config_base = 'icecast.xml'
    _config = dict(
        location='Neptune',
        admin_email='sushi@trident.sea',
        password='hack-it-up',
        hostname='127.0.0.1',
        port='8777',
    )

    async def _prepare(self, config):
        """Get the configuration file ready."""
        self._command.extend(['-c', str(config)])

    @classmethod
    def client(cls, mount):
        """Return a process that streams to the icecast server."""
        cmd = 'ffmpeg'
        uri = 'icecast://source:{}@{}:{}/{}'.format(
            cls._config['password'],
            cls._config['hostname'],
            cls._config['port'],
            mount
        )
        # Maybe:
        # -reconnect 1 -reconnect_at_eof 1 -reconnect_streamed 1 \
        # -reconnect_delay_max 2000
        flags = [
            '-re',
            '-ac', '2',
            '-ar', '44.1k',
            '-f', 's16le',
            '-i', '-',
            '-vn',
            '-codec:a', 'libvorbis',
            '-q:a', '8.0',
            '-content_type', 'audio/ogg',
            '-f', 'ogg',
            uri
        ]
        return Spool(cmd, xflags=flags)
